/*
   cmd.c -- PIKT piktc command-line processing routines

   $Header: /var/cvs/PIKT/pikt/src/piktc/cmd.c,v 1.57 2007/01/18 13:23:39 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"
#include "piktc.h"

nam *expandhosts(nam *);
nam *expanditems(nam *);
/*
int chkitems(nam *, nam *);
 */

/* check whether alerts, programs, files, objectsets & hosts specified in
   the command line are registered in the config files; also whether
   command/alert/program/files/objectset/host specifications are consistent
   with other command-line options; note that cmdparse() ensures that "all",
   if present, is the sole member of a list
 */
void
preprocesscmd (void)
{
	static char fnnam[] = "preprocesscmd";
	nam *nmp1;
	ident *idp;

	if (    ishelp
	     || isshow_version
	     || (! havecmdhosts)
	     || ((ischeck || isdoublecheck || islist_elements) && havecmditems)
	     || ((! ischeck && ! isdoublecheck && ! islist_elements && ! isinstall_piktconf && ! isrestart_piktd && ! isrestart_piktc_svc && ! iskill_piktd && ! iskill_piktc_svc && ! islist_items) && (! havecmditems))
	     || ((isdisable || isenable || isshow_status) && (! incmdalerts && ! excmdalerts))
	     || ((isexec || isexecwait) && (! incmdcommands && ! incmdalerts && ! excmdalerts))
	     || (isexpand_macros && ! incmdcommands)
	     || ((isdelete || iscksum || isdiff || isview_actual || isview_auth || isinstall || islist_files) && ! havecmditems)
	   ) {
		bailout = TRUE;
		return;
	}
	/* nothing to do here
	foreach (nmp1, incmdcommands)
		;
	 */
	if (! bailout) {
		commandsall = TRUE;
		cmdcommands = incmdcommands;
	}
	foreach (nmp1, incmdhosts) {
		if (hostsall)
			break;
		if (! (idp = lookupident(nmp1->name, SYSCFG))) {
			logreturn(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "unknown os, host, group, or alias: %s", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		} else if ((idp->type != OPS) && (idp->type != GRP) && (idp->type != HST) && (idp->type != ALS)) {
			logreturn(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s is not an os, group, host, or alias", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	foreach (nmp1, excmdhosts) {
		if (hostsnone)
			break;
		if (! (idp = lookupident(nmp1->name, SYSCFG))) {
			logreturn(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "unknown os, host, group, or alias: %s", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		} else if ((idp->type != OPS) && (idp->type != GRP) && (idp->type != HST) && (idp->type != ALS)) {
			logreturn(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s is not an os, group, host, or alias", nmp1->name);
			nmp1->state = ERROR;
			bailout = TRUE;
		}
	}
	if (! bailout)
		cmdhosts = purge(cfghosts, incmdhosts, excmdhosts, HST);
	/* no, if no hostset results from the interplay of +H and -H settings,
	   let's not signal a piktc_usage() call by setting bailout to TRUE
	if (! cmdhosts)
		bailout = TRUE;
	 */
	return;
}

/* register the command-line name or command in the specified list
 */
nam *
regname (nam **list, char *name, int command)
{
	static char fnnam[] = "regname";
	char *cp;
	nam *nmp, *nnmp;

	if (! command) {
		/* look to see if this is a file path;
		   if so, disregard path for now
		 */
		if ((cp = strrchr(name, '/'))) {
			name = cp+1;
		}
		/* look to see if this name is already registered;
		   note: it is perfectly okay to repeat a command
		 */
		if (*list) {
			foreach (nmp, *list) {
				if (! strcmp(nmp->name, name))
					return nmp;
			}
		}
	}
	/* not already registered, or is command, so proceed */
	nnmp = (nam *) mymalloc(sizeof(nam), fnnam);
	nnmp->next = (nam *) NULL;
	nnmp->name = mystrdup(name, fnnam);
	nnmp->aux = (char *) NULL;
	nnmp->dir = (char *) NULL;
	nnmp->file = (char *) NULL;
	nnmp->mode = (char *) NULL;
	nnmp->filuid = (char *) NULL;
	nnmp->filgid = (char *) NULL;
	nnmp->state = ADD;
	/* add to list */
	if (*list) {
		for (nmp=*list; nmp->next; nmp=nmp->next) ;
		nmp->next = nnmp;
	} else
		*list = nnmp;
	return *list;	
}

/* garbage collect list items
 */
void
retire (nam **names)
{
	nam *nmp;
	static nam *tnmp;

	nmp = *names;
	while (nmp) {
		tnmp = nmp;
		nmp = nmp->next;
		FREE(tnmp->name);
		FREE(tnmp->aux);
		FREE(tnmp->dir);
		FREE(tnmp->file);
		FREE(tnmp->mode);
		FREE(tnmp->filuid);
		FREE(tnmp->filgid);
		FREE(tnmp);
	}
	*names = (nam *) NULL;
	return;
}

/* do a lookup on the names specified in the command line
 */
nam *
lookupname (nam *list, char *name)
{
	char *cp;
	static nam *nmp;

	/* if name is a filepath, disregard path */
	if ((cp = strrchr(name, '/')))
		name = cp+1;
	foreach (nmp, list) {
		if (! strcmp(name, nmp->name))
			break;
	}
	return nmp;
}


/* given an include list, purge members in the given exclude list
 */
nam *
purge (nam *cfgitems, nam *incmditems, nam *excmditems, int type)
{
	nam *initems, *exitems;
	static nam *nmp1, *nmp2;

	initems = exitems = (nam *) NULL;
	if (incmditems) {
		if (! strcmp(incmditems->name,"all")) {
			initems = cfgitems;
		} else if (type == HST) {
			initems = expandhosts(incmditems);
			if (! initems)
				return (nam *) NULL;
		} else if (havecmdmacros[type]) {
			initems = expanditems(incmditems);
			if (! initems)
				return (nam *) NULL;
		} else {
			initems = incmditems;
		}
	}
	if (! initems) {
		initems = cfgitems;
	}
	if (excmditems) {
		if (! strcmp(excmditems->name,"all")) {
			exitems = cfgitems;
		} else if (type == HST) {
			exitems = expandhosts(excmditems);
		} else if (havecmdmacros[type]) {
			exitems = expanditems(excmditems);
		} else {
			/* note the inconsistency here with how initems
			   is treated above; this is because exitems is
			   (potentially) retire()'d below, while initems
			   is never retire()'d; if we simply set
			   exitems = excmditems, and if we retire() exitems
			   below, we also blow away excmditems, and
			   the program segfaults at the second host
			   go-around; so we must in effect replicate the
			   excmditems list by invoking expanditems(), and
			   it is this replicated list that is retired
			 */
			exitems = expanditems(excmditems);
		}
		/* note: exitems might now be NULL; this makes necessary
		   checking if exitems is NULL before calling retire(&exitems)
		   below
		 */
	}
	/* foreach of the items in the exclude list, if any,
	   seek out and purge from the include list
	 */
	foreach (nmp1, exitems) {
		if (! (nmp2 = initems))	/* if excluded initems down to nothing */
			break;
		if (! strcmp(nmp1->name, nmp2->name)) {
			initems = nmp2->next;
		} else {
			while (nmp2->next) {
				if (! strcmp(nmp1->name, nmp2->next->name)) {
					nmp2->next = nmp2->next->next;
					break;
				}
				nmp2 = nmp2->next;
			}
		}	
	}
	if (excmditems && strcmp(excmditems->name, "all") && exitems)
		retire(&exitems);
	return initems;
}

/* given a host/alias/group list, replace any host alias with the hostname
   it represents, and replace any group with the hostnames it represents;
   e.g., the list

      vienna rome db linux

   would expand to

      vienna milan athens2 athens4 milan seville plymouth

   if vienna were a plain hostname; rome were an alias for milan; db were
   a host group including athens2, athens4, and milan; and linux were an
   os group including seville and plymouth
 */
nam *
expandhosts (nam *oldlist)
{
	static char fnnam[] = "expandhosts";
	nam *newlist;
	nam *nmp;
	ident *idp;
	unsigned long *hbs;
	unsigned long hbsl;
	register int i, j;

	newlist = (nam *) NULL;
	foreach (nmp, oldlist) {
		idp = lookupident(nmp->name, SYSCFG);
		switch (idp->type) {
			case HST:
				regname(&newlist, idp->name, FALSE);
				break;
			case ALS:
				regname(&newlist, idp->aux, FALSE);
				break;
			case GRP:
			case OPS:
				hbs = idp->bs;
				/* foreach unsigned long in the host bitset */
				for (i=0; i<hbslongs; i++) {
					hbsl = *(hbs+i);
					/* foreach bit in the current
					   unsigned long
					 */
					for (j=0; j<8*sizeof(unsigned long); j++) {
						if (hbsl & (1L<<j))
							regname(&newlist, *(hosts+((i*sizeof(unsigned long)*8)+j)), FALSE);
					}
				}
				break;
			default:
				logreturn(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s is not an os, group, or host", nmp->name);
				bailout = TRUE;
				return (nam *) NULL;
		}
	}
	return newlist;
}

/* given a command-line items list, expand its components; e.g., the
   macro item

   +F =sshcfg

   might become

   +F ssh_config sshd_config
 */
nam *
expanditems (nam *oldlist)
{
	static char fnnam[] = "expanditems";
	nam *newlist;
	nam *nmp;
	static char *list;
	static char *item;
	static char *itm;
	register int i;
	register char c;

	newlist = (nam *) NULL;
	foreach (nmp, oldlist) {
		list = item = expand(nmp->name, NONE);
		i = 0;
		while ((c = *(list+i))) {
			if (isspace((int)c)) {
				*(list+i) = '\0';
				itm = mystrdup(item, fnnam);
				regname(&newlist, itm, FALSE);
				item = list+i+1;
			}
			i++;
		}
		itm = mystrdup(item, fnnam);
		regname(&newlist, itm, FALSE);
		FREE(list);
	}
	return newlist;
}

/*
   check command-line items against the config files for discrepancies
 */
/*
int
chkitems (nam *cmditems, nam *cfgitems)
{
	static char fnnam[] = "chkitems";
	nam *nmp1, *nmp2;

	foreach (nmp1, cmditems) {
		if (! strcmp(nmp1->name,"all"))
			break;
		foreach (nmp2, cfgitems)
			if (! strcmp(nmp1->name, nmp2->name))
				break;
		if (! nmp2) {
			logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "unknown identifier: %s", nmp1->name);
			nmp1->state = ERROR;
			return FALSE;
		}
	}
	return TRUE;
}
*/
